#include "Motor.h"

Motor::Motor(int8_t in1, int8_t in2, int8_t pwm, int8_t enc1, int8_t enc2) {
  _in1 = in1;
  _in2 = in2;
  _pwm = pwm;
  pinMode(_in1, OUTPUT);
  pinMode(_in2, OUTPUT);
  pinMode(_pwm, OUTPUT);

  _enc1 = enc1;
  _enc2 = enc2;
  pinMode(_enc1, INPUT_PULLUP);
  pinMode(_enc2, INPUT_PULLUP);
}

double Motor::getSpeed() {
  return _current_speed;
}

double Motor::_calcSpeed() {
  int startTime = _history[_historyLast].time;
  int startTicks = _history[_historyLast].data;
  int sum = 0, items = 0, l;
  float dt;
  for (int i = 0; i < 5; i++) {
    l = (i + _historyLast + 1) % 5;
    if (i == 4) _history[l].time = micros();
    if (_history[l].actual && _history[l].time - startTime != 0) {
      dt = (_history[l].time - startTime) / 1000000.0;
      sum += (_history[l].data - startTicks) / dt;
      items++;
    }
  }
  if (items == 0) return 0;
  sum /= items;
  return sum;
}

int Motor::getTicks() {
  return -_ticks;
}

void Motor::tick() {
  _ticks -= (digitalRead(_enc1) == HIGH && digitalRead(_enc2) == LOW) * 2 - 1;
  _history[_historyLast].time = micros();
  _history[_historyLast].data = _ticks;
  _history[_historyLast].actual = true;
  _historyLast = (_historyLast + 1) % 5;
}

void Motor::process() {
  double v = _calcSpeed();
  _current_speed = v;//0.25 * _current_speed + 0.5 * v + 0.25 * _current_speed_prev;
  _current_speed_prev = v;
  double dt = (float)(millis() - _tmr) / 1000.0;
  _tmr = millis();
  int power = _power;
  double err, d, speed = getSpeed();
  // prevent ticks overload int type
  /*if (abs(_ticks) > 2147483647 - speed * 10) {
    int delta = abs(2147483647 - abs(_ticks));
    if (_ticks < 0) delta *= -1;
    _ticks -= delta;
    for (int i = 0; i < 5; i++) {
      _history[i].data -= delta;
    }
  }*/
  if (_speed_mode == 1) {
    err = _speed - speed;
    _i += err * dt;
    power = constrain(err * kp + _i * ki, -255, 255);
  }
  if (_target_mode == 1) {
    err = _target - _ticks;
    //if (err == 0) _ti = 0;
    //   if ((abs(err) < abs(speed)) || (_power < 0 && err > 0) || (_power > 0 && err < 0) || (abs(err) > 0 && _power == 0)) {
    _ti += err * dt;
    d = (err - _tspe) / dt;
    _tspe = err;
    power = err * tkp + _ti * tki + d * tkd;
    //  }
  }

  set(power);
}

void Motor::setPower(int power) {
  _power = power;
  _speed_mode = 0;
}

void Motor::setSpeed(double speed) {
  _speed = speed;
  _speed_mode = 1;
}

void Motor::setTarget(double degrees) {
  _target = degrees / 360.0 * redution * ticksPerRevolution;
  _target_mode = 1;
}

void Motor::removeTarget() {
  _target_mode = 0;
}

void Motor::resetEncoder() {
  _ticks = 0;
}

void Motor::set(int value) {
  value = constrain(value, -255, 255);
  if (value >= 0) {
    value = map(value, 0, 255, minSpeed, maxSpeed);
    digitalWrite(_in1, HIGH);
    digitalWrite(_in2, LOW);
  } else {
    value = map(-value, 0, 255, minSpeed, maxSpeed);
    digitalWrite(_in1, LOW);
    digitalWrite(_in2, HIGH);
  }
  analogWrite(_pwm, value);
}