#ifndef Motor_h
#define Motor_h
#include <Arduino.h>

struct EcoderStamp {
  int time; // millis
  int data; // ticks
  bool actual = false;
};

class Motor {
  public: 
    Motor(int8_t in1, int8_t in2, int8_t pwm, int8_t enc1, int8_t enc2);
    void process();
    void setPower(int power);
    void setSpeed(double speed);
    void setTarget(double degrees);
    void removeTarget();
    int getTicks();
    void resetEncoder();
    double getSpeed();
    void set(int value);

    // init interrupts by yourself and for each hall sensor call tick()
    void tick();

    int minSpeed = 0;
    int maxSpeed = 255;
    int ticksPerRevolution = 12;
    double redution = 28.6;
    double wheelRad = 25;

    double kp = 2, ki = 10;
    double tkp = 10, tki = 0, tkd = 1;
  private:
    double _calcSpeed();

    // config
    int8_t _in1, _in2, _pwm, _enc1, _enc2;
    // settings
    int _speed_mode = 0; // 0 power 1 speed
    int _target_mode = 0; // 0 targetless 1 target 
    int _power = 0;
    double _speed = 0;
    int _target = 0;
    double _current_speed = 0;
    double _current_speed_prev = 0;

    // calculation variables
    int _ticks = 0;
    bool _tickCommited = true;
    EcoderStamp _history[5];
    int _historyLast = 0;
    long long _tmr = 0;
    double _i = 0; // integral
    double _ti = 0, _tspe = 0; // target integral, target speed previous error
};

#endif