#include "motor.h"
#include "I2Cdev.h"
#include "MPU6050_6Axis_MotionApps20.h"
const float Pi = 3.14159;
Motor right(6, 9, 11, 3, 5);
Motor left(8, 7, 10, 2, 4);

// gyro
MPU6050 mpu;
#define BUFFER_SIZE 100
uint8_t fifoBuffer[45];
float origin = 0;
float gyro = 0;
float zero = 0;

// odometry
#define FORWARD_SPEED 140
#define ROTATE_SPEED 160
int speed = 0;
float target = 0;  // avarage ticks
float odoK = 1;    // left ticks / right ticks
bool side = true;
bool stoped = false;
unsigned long stopedAt = 0;

void ltick() {
  left.tick();
}
void rtick() {
  right.tick();
}

int dir = 0;
int x = 0, y = 0;
double wheelBase = 70;
double wheelDiam = 26;

void setup() {
  attachInterrupt(digitalPinToInterrupt(2), ltick, RISING);
  attachInterrupt(digitalPinToInterrupt(3), rtick, RISING);

  pinMode(A0, INPUT);
  pinMode(A1, INPUT);
  Serial.begin(9600);
  // Wire.begin();
  //Wire.setClock(1000000UL);   // разгоняем шину на максимум
  // инициализация DMP
  // mpu.initialize();
  // calibration();
  // mpu.dmpInitialize();
  // mpu.setDMPEnabled(true);
  // rotate(-90);
}

void rotate(double angle) {
  target = abs(angle / 360.0) * (wheelBase / wheelDiam) * 28.6 * 12.0;
  odoK = -1;
  side = angle > 0;
}

void forward(double dist) {
  odoK = 1;
  target = (dist / (wheelDiam * Pi)) * 28.6 * 12.0;
}

bool doRight = false;
unsigned long last = millis();

bool moveSide = false;
int leftI = 0;
int rightI = 0;

bool doFront = false;
unsigned long doFrontAt = 0;

void actionEnd() {
  if (x == 2 && y == 2) {
    left.set(0);
    right.set(0);
    return;
  }
  if (!stoped) {
    stoped = true;
    stopedAt = millis();
    return;
  }

  stoped = false;
  left.resetEncoder();
  right.resetEncoder();
  moveSide = true;
  bool rightWall = analogRead(A1) < 600;
  bool frontWall = analogRead(A0) < 700;

  if (doRight) {
    doRight = false;
    forward(160);
    if (dir == 0) y++;
    else if (dir == 1) x++;
    else if (dir == 2) y--;
    else if (dir == 3) x--;
    return;
  }
  if (doFront) {
    doFront = false;
    rotate(-90);
    dir--;
    if (dir < 0) dir = 3;
    return;
  }
  if (!rightWall) {
    doRight = true;
    rotate(90);
    dir++;
    if (dir == 4) dir = 0;
    return;
  }
  if (frontWall) {
    doFront = true;
    doFrontAt = millis();
    return;
  }

  forward(160);
  if (dir == 0) y++;
  else if (dir == 1) x++;
  else if (dir == 2) y--;
  else if (dir == 3) x--;
}

void loop() {
  if (doFront) {
    double v = sqrt(analogRead(A0));
    double err = v - 7;
    if (millis() - doFrontAt < 500) {
      double u = 0;
      if (err > 0) u = err * 5;
      else u = err * 50;
      left.set(u);
      right.set(u);
    } else {
      actionEnd();
    }
    return;
  }
  if (odoK > 0) {
    if (stoped) {
      int err = target * 2 - left.getTicks() - right.getTicks();
      double u = err * 0.5;
      left.set(u);
      right.set(u);
      if (millis() - stopedAt > 500) return actionEnd();
    } else {
      if (left.getTicks() + right.getTicks() < target * 2) {
        double u = 0;
        if (analogRead(A1) < 700) {
          double s = sqrt(analogRead(A1));
          double err = 4.0 - s;
          if (err > 0) {
            u = err * 50;
          } else u = err * 4;
          moveSide = true;
        } else {
          if (moveSide) {
            leftI = left.getTicks();
            rightI = right.getTicks();
            moveSide = false;
          }
          u = ((left.getTicks() - leftI) - (right.getTicks() - rightI)) * 4;
        }
        left.set(FORWARD_SPEED - u);
        right.set(FORWARD_SPEED + u);
        if (pow(analogRead(A0), 2) < 49) return actionEnd();
      } else return actionEnd();
    }
    double v = sqrt(analogRead(A0));
    if (v < 7) {
      double err = v - 7;
      double u = err * 50.0;
      left.set(u);
      right.set(u);
    }
  } else {
    if (side) {
      if (stoped) {
        left.set(target - left.getTicks());
        right.set(-right.getTicks() - target);
        if (millis() - stopedAt > 500) actionEnd();
      } else {
        if (left.getTicks() - right.getTicks() < target * 2) {
          int err = left.getTicks() + right.getTicks();
          double u = err;
          left.set(max(0, ROTATE_SPEED - u));
          right.set(min(0, -(ROTATE_SPEED + u)));
        } else actionEnd();
      }
    } else {
      if (stoped) {
        left.set(-left.getTicks() - target);
        right.set(target - right.getTicks());
        if (millis() - stopedAt > 500) actionEnd();
      } else {
        if (-left.getTicks() + right.getTicks() < target * 2) {
          int err = -left.getTicks() - right.getTicks();
          double u = err;
          left.set(-(ROTATE_SPEED - u));
          right.set(ROTATE_SPEED + u);
        } else actionEnd();
      }
    }
  }
}

void updateGyro() {
  static uint32_t tmr;
  if (millis() - tmr < 11) return;
  if (!mpu.dmpGetCurrentFIFOPacket(fifoBuffer)) return;
  // переменные для расчёта (ypr можно вынести в глобал)
  Quaternion q;
  VectorFloat gravity;
  float ypr[3];
  // расчёты
  mpu.dmpGetQuaternion(&q, fifoBuffer);
  mpu.dmpGetGravity(&gravity, &q);
  mpu.dmpGetYawPitchRoll(ypr, &q, &gravity);
  origin = degrees(ypr[0]);
  gyro = origin - zero;
  tmr = millis();  // сброс таймера
}

void calibration() {
  long offset;
  long offsetOld;
  int16_t mpuGet[6];
  mpu.setFullScaleAccelRange(MPU6050_ACCEL_FS_2);
  mpu.setFullScaleGyroRange(MPU6050_GYRO_FS_250);
  mpu.setZGyroOffset(0);
  delay(10);
  for (byte n = 0; n < 10; n++) {
    offset = 0;
    for (byte i = 0; i < BUFFER_SIZE; i++) {
      mpu.getMotion6(&mpuGet[0], &mpuGet[1], &mpuGet[2], &mpuGet[3], &mpuGet[4], &mpuGet[5]);
      offset += (long)mpuGet[5];
    }
    offset = offsetOld - ((long)offset / BUFFER_SIZE); 
    offsetOld = offset;
    mpu.setZGyroOffset(offset / 4);
    delay(10);
  }
}